# Introduction

This document provides an overview of the TIBCO Intelligent Equipment Accelerator.  

# Overview

## Business Scenario

Most modern equipment are instrumented in some way with a variety of telemetry captured from sensors, from cars to electronics to lightbulbs. Gathering this data and making sense of it all is a key problem for equipment owners. Once data is captured either on edge devices or within a core infrastructure, it then becomes a challenge to detect patterns and meaningful behaviours from the noise. Through the use of rule-based systems and data science models, actionable insights can be gleaned. That allows the ability to take action in developing situations, or just capture the data to refine models for future improvements to the system.

## Concepts

The Intelligent Equipment Accelerator is based on a configuration-driven, generic data model. At the top level there are two main concepts:

**Devices** -- are anything that produce a stream of data. Also known as sensors. Typically produce data triplets at high frequency, consisting of a unique identifier, a timestamp, and a data value. Devices are attached to a single Entity, but an Entity can have multiple Devices.

**Entities** -- are anything else. This can be factories, production lines, equipment, aircraft, buses, ovens, drilling rigs... anything. Organized into hierarchies, one Entity may have a single parent, but multiple children.

To help with configuration, the Accelerator also separates configuration into  Templates and Instances. 

**Instances** -- are physical example of a Device or Entity, equivalent in object-oriented programming to an Object Instance. They are linked to a single Template, have a physical location, and a unique identifier like a serial number.

**Templates** -- definition of common properties for all Instances of a given Template, equivalent in object-oriented programming to a Class. May also be known as a type. Will not have a physical location or a unique identifier like a serial number (but could be a unique model number).

Since Devices often send only single data points at a time, it is often useful to aggregate these together into virtual rows of data for processing.

**Features** -- are linked to a single Device Instance or to a Device Template associated with an Entity Template. Represents a single value in a virtual row.

**Feature Sets** -- are logical groupings of Features into a single virtual row. This virtual row can then be passed to rules and data science models to evaluate multi-variate conditions and states.

This configuration looks like this:

![](./doc/images/model-hierarchy.png)

In addition, users can configure Modules, which link to physical EventFlow application modules implementing specific business rules. These may be implemented as Validation Modules, Cleansing Modules, or Rule Modules. These modules are then linked to Devices, Device Templates, and Feature Sets so they are called during the processing of data from these data sources. 

The Accelerator captures data feeds from external systems as reports. 

**Alert** -- reports from external systems of alert conditions.

**Reading** -- device reading reports consisting of a triplet of unique identifier, reading date and time, and the value.

**Status** -- a condition status for a given Device or Entity. For example, a production line or pump operating status.

**Part** -- a part produced report used as part of operational metrics.

**Position** -- a physical location for a given Entity that may change over time.

After processing the inbound reports, the Accelerator produces external actions.

**Alerts** -- similar to Alert reports, indicates an alert condition on a Device or Entity.

**Status** -- similar to Status reports, indicates the status of a given Device or Entity has changed.

**Readings** -- certain rules may produce additional Feature values as part of the rule execution. For example, an autoencoder rule may generate a cluster number and a reconstruction error. These Feature values are produced as new Readings.

The dynamic data model looks like this:

![](./doc/images/model-dynamic.png)


## Benefits and Business Value

Most modern equipment today are instrumented with some sort of sensor. We can use the streaming data from these sensors, combined with context information from various systems to gain a complete real-time view of all operations in order to rapidly resolve current issues and intervene to address preventable problems before they occur.

## Technical Scenario

The Accelerator demonstrates real-time monitoring and alerting for several different use cases.

**ESP** -- This use case shows how field equipment can be monitored in oil production. Electric submersible pumps (ESPs) are down-hole pump equipment inside oilwells. By analyzing historical data, a failure pattern was detected that indicates likely pump failure when there is a rise in intake pressure correlating with a drop motor current draw. By detecting this pattern and alerting, this gives field operators the opportunity to schedule downtime for pump maintenance rather than waiting for the pump to fail. The Accelerator is configured with a simple hierarchy of ESPs and both pressure and motor amps devices. A recorded dataset can be replayed which will result in raised alerts.  

**Heavy Equipment** -- Mobile heavy equipment monitoring is illustrated with this use case. A fleet of equipment operating in a mine are monitored in real-time for engine signals. When defined patterns occur then alerts are raised for preventative maintenance activities such as spark plug replacements. The location of the equipment is also tracked in real-time and displayed on a dashboard.

**Power Plant** -- In this use case a set of gas-fired power plants across Europe are monitored for emissions profiling through multivariate analysis and data science models. A simple k-means clustering model and more sophisticated autoencoder model classify the state of the power plant based on multiple metrics and certain states indicate emissions out of allowable range. A recorded dataset can be replayed to trigger the model classifications and alerts.

**Servers** -- This use case shows IT infrastructure monitoring with various implementation layers including physical infrastructure, software platform, and service. When parameters go out of bound then alerts are raised to indicate possible environment instability.

**Widgets** -- Real-time monitoring of manufacturing is illustrated by this use case. A hierarhcy of factories, production lines, and equipment are monitored for equipment failures and efficiency. It monitors part production and calculates overall equipment effectiveness in real-time. Equipment sensors are monitored and alert rules detect out of range readings. A simulator generates random sensor data and part production data.

## Components

At the heart of the Accelerator is the Event Manager which is implemented using TIBCO Streaming. This is a series of EventFlow applications that receive reports from various sources and perform the required business logic. The Event Manager includes extension points for validations, cleansing, rules, and statistics that allow users to configure their own logic modules and attach them to the data stream. 

As data flows through the system, it is passed through to a real-time datamart implemented using Live Datamart and displayed on a real-time dashboard implemented using Spotfire®. System metadata is stored in a master data repository implemented using EBX®, which allows modification using the product UI. A custom web user interface is also provided for metadata editing.

![](./doc/images/component.png)

# Installation

Before installing the Accelerator, the user will need to decide on a deployment environment:

Environment | Description
------- | -------
[Bare metal or virtual machine](./doc/quick-start.md) | Suitable for development, demo, and runtime uses. Requires appropriate hardware, base operating system, and database configured and running.
[Hybrid cloud](./doc/cloud-getting-started.md) | Suitable for demo and runtime uses. Requires TIBCO Cloud account with enabled subscriptions, and an AWS account with access to EKS.

For further instructions how to setup these environments select the appropriate link above.
