/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

// Import required resources
var createError = require('http-errors');
var express = require('express');
var logger = require('morgan');
var path = require('path');
var properties = require('properties-reader');
var proxy = require('express-http-proxy');

// Create the application
var app = express();

// setup routers
var environmentRouter = require('./routes/environment');
var webappRouter = require('./routes/webapp');
var usersRouter = require('./routes/users');
const { resolve } = require('path');

// Environment properties
var cicProps = properties('src/webapp/cic.properties');

// view engine setup
app.set('views', path.join(__dirname, 'views'));
app.set('view engine', 'pug');

app.use(logger('dev'));
app.use(express.json());
app.use(express.urlencoded({extended: false}));
app.use(express.static(path.join(__dirname, '../public')));


// --------------------------------------------------------------------------------
// PROXY

// Setup proxy for CIC Account
app.set('trust proxy', 1);

// Proxy for login because this requires extra validation for authorization
app.use('/service/account/us/login', proxy(cicProps.get('us.TIBCO_CLOUD_ACCOUNT'), {
  proxyReqPathResolver: function (req) {
    return '/idm/v3/login-oauth';
  },
  proxyReqOptDecorator: function(proxyReqOpts, srcReq) {
    delete proxyReqOpts.headers['origin'];
    return proxyReqOpts;
  },
  userResHeaderDecorator(headers, userReq, userRes, proxyReq, proxyRes) {  
    headers['set-cookie'] = '';
    return headers;
  },
  userResDecorator: function(proxyRes, proxyResData) {
    // TODO validate authorization
    return proxyResData;
  }
}));

// Proxy for all other US CIC account service calls
app.use('/cic/account/us', proxy(cicProps.get('us.TIBCO_CLOUD_ACCOUNT'), {
  proxyReqPathResolver: function (req) {
    return req.path;
  },
  proxyReqOptDecorator: function(proxyReqOpts, srcReq) {
    delete proxyReqOpts.headers['origin'];
    return proxyReqOpts;
  },
  userResHeaderDecorator(headers, userReq, userRes, proxyReq, proxyRes) {  
    headers['set-cookie'] = '';
    return headers;
  },
  userResDecorator: function(proxyRes, proxyResData) {
    return proxyResData;
  }
}));

// Proxy for login because this requires extra validation for authorization
app.use('/service/account/eu/login', proxy(cicProps.get('eu.TIBCO_CLOUD_ACCOUNT'), {
  proxyReqPathResolver: function (req) {
    return '/idm/v3/login-oauth';
  },
  proxyReqOptDecorator: function(proxyReqOpts, srcReq) {
    delete proxyReqOpts.headers['origin'];
    return proxyReqOpts;
  },
  userResHeaderDecorator(headers, userReq, userRes, proxyReq, proxyRes) {  
    headers['set-cookie'] = '';
    return headers;
  },
  userResDecorator: function(proxyRes, proxyResData) {
    console.log(proxyRes);
    // TODO validate authorization
    return proxyResData;
  }
}));

// Proxy for all other EU CIC account service calls
app.use('/cic/account/eu', proxy(cicProps.get('eu.TIBCO_CLOUD_ACCOUNT'), {
  proxyReqPathResolver: function (req) {
    return req.path;
  },
  proxyReqOptDecorator: function(proxyReqOpts, srcReq) {
    delete proxyReqOpts.headers['origin'];
    return proxyReqOpts;
  },
  userResHeaderDecorator(headers, userReq, userRes, proxyReq, proxyRes) {  
    headers['set-cookie'] = '';
    return headers;
  },
  userResDecorator: function(proxyRes, proxyResData) {
    return proxyResData;
  }
}));


// --------------------------------------------------------------------------------
// ENVIRONMENT

// Setup environment router, no validation of cookies for these services
app.use('/', environmentRouter);

// Validate required cookies exist
function cookiesExists(req, names) {
  return true;
};

// Setup validation for cookies
app.use(function(req, res, next){
  if(cookiesExists(req, ['tsc', 'domain'])) {
    next();
  }
  else {
    next(createError(404));
  }
});

// --------------------------------------------------------------------------------
// OTHER ROUTERS

// All subsequent services require cookies to be set to work properly
// Setup routers
app.use('/', webappRouter);
app.use('/users', usersRouter);


// --------------------------------------------------------------------------------
// 404 and ERROR

// catch 404 and forward to error handler
app.use(function(req, res, next) {
  next(createError(404));
});

// error handler
app.use(function(err, req, res, next) {
  // set locals, only providing error in development
  res.locals.message = err.message;
  res.locals.error = req.app.get('env') === 'development' ? err : {};

  // render the error page
  res.status(err.status || 500);
  res.render('error');
});


// --------------------------------------------------------------------------------
// EXPORT

// Export the application module
module.exports = app;

// Port is being set in bin/www
//app.listen(3000);