/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

var axios = require('axios');
var express = require('express');
var fs = require('fs');
var properties = require('properties-reader');
const { route } = require('./users');
var Simulator = require('../webapp/simulator_kafka')
var Provisioning = require('../webapp/provisioning_spotfire');

// Create the router
var router = express.Router();

// Environment properties
var cicProps = properties('src/webapp/cic.properties');

// Start page for redirects
var START_PAGE = 'app/iot/index.html';

// Simulator singletom
var simulator = null;


// --------------------------------------------------------------------------------
// ROOT
router.get('/', function(req, res, next) {
  res.redirect(301, START_PAGE);
  //res.render('index', { title: 'Express' });
});


// --------------------------------------------------------------------------------
// CONTEXT
// Read the configuration file
router.get('/service/context/configuration', function(req, res, next) {
  fs.readFile(__dirname + '/../../data/context/configuration.json', function (err, data) {
    if (err) {
      throw err; 
    }
    let response = JSON.parse(data.toString());
    res.json(response);
  });
});

// Write the configuration file
router.post('/service/context/configuration', function(req, res, next) {
  var payload = JSON.stringify(req.body, null, 2);
  fs.writeFile( __dirname + '/../../data/context/configuration.json', payload, function (err, data) {
    if (err) {
      throw err; 
    }
    res.json("{}");
  });
});


// --------------------------------------------------------------------------------
// OAUTH

// Retrieve TCM roles
router.get('/service/cic/oauth/messaging/roles/:region', function(req, res, next) {
  var region = req.params.region;
  if(region == null) {
    res.status(500).send("Invalid region");
    return;
  }

  var url = '/tcm/ui/uiapi/v1/roles';
  var baseUrl = cicProps.get(region + '.TIBCO_CLOUD_MESSAGING');

  axios({
    url: url,
    method: 'GET',
    baseURL: baseUrl,
    headers: {
      'Authorization': req.headers['authorization']
    }
  })
  .then(function(response){
    res.status(200).send(response.data);
  })
  .catch(function(err){
    res.status(err.response.status).send(err);
  });

});

// Retrieve TCM role configuration
router.get('/service/cic/oauth/messaging/roleconfig/:region/:role', function(req, res, next) {
  var region = req.params.region;
  if(region == null) {
    res.status(500).send("Invalid region");
    return;
  }

  var role = req.params.role;
  if(role == null) {
    res.status(500).send("Invalid role");
    return;
  }

  var url = '/tcm/ui/uiapi/v1/roles/' + role + '/config/channel';
  var baseUrl = cicProps.get(region + '.TIBCO_CLOUD_MESSAGING');

  axios({
    url: url,
    method: 'GET',
    baseURL: baseUrl,
    headers: {
      'Authorization': req.headers['authorization']
    }
  })
  .then(function(response){
    res.status(200).send(response.data);
  })
  .catch(function(err){
    res.status(err.response.status).send(err);
  });

});

// Retrieve TCM channels
router.get('/service/cic/oauth/datastreams/channels/:region', function(req, res, next) {
  var region = req.params.region;
  if(region == null) {
    res.status(500).send("Invalid region");
    return;
  }

  var url = '/ui/api/v1/plugins/tcm/channels';
  var baseUrl = cicProps.get(region + '.TIBCO_CLOUD_DATASTREAMS');

  axios({
    url: url,
    method: 'GET',
    baseURL: baseUrl,
    headers: {
      'Authorization': req.headers['authorization']
    }
  })
  .then(function(response){
    res.status(200).send(response.data);
  })
  .catch(function(err){
    res.status(err.response.status).send(err);
  });

});

// Retrieve TCE sandboxes
router.get('/service/cic/oauth/events/sandboxes/:region', function(req, res, next) {
  var region = req.params.region;
  if(region == null) {
    res.status(500).send("Invalid region");
    return;
  }

  var url = '/tceorch/v1/sandboxes';
  var baseUrl = cicProps.get(region + '.TIBCO_CLOUD_EVENTS');
  
  axios({
    url: url,
    method: 'GET',
    baseURL: baseUrl,
    headers: {
      'Authorization': req.headers['authorization']
    }
  })
  .then(function(response){
    res.status(200).send(response.data);
  })
  .catch(function(err){
    res.status(err.response.status).send(err);
  });

});

// Download DXP
router.get('/service/spotfire/analysis/:region/:oauth', async function(req, res, next) {
  var region = req.params.region;
  var oauth = "Bearer " + req.params.oauth;
  var provisioner = new Provisioning.Provisioner(); 
  provisioner.download(region, oauth, res);
});


// --------------------------------------------------------------------------------
// SIMULATION
// Get testcases
router.get('/service/simulation/testcases', function(req, res, next) {
  const testFolder = __dirname + '/../../data/testcases';
  fs.readdir(testFolder, function(err, files) {
    let testcases = [];
    files.forEach(function(file) {
      let testcase = JSON.parse(fs.readFileSync(testFolder + "/" + file).toString());
      testcases.push(testcase);
    });

    let response = {
      "testcases": testcases
    }
    res.json(response);
  });
});  

// Subscribe to simulation status event stream
router.get('/service/simulation/status', async function(req, res, next) {
  res.set({
    'Cache-Control': 'no-cache',
    'Content-Type': 'text/event-stream',
    'Connection': 'keep-alive'
  });
  res.flushHeaders();

  // Tell the client to retry every 10 seconds if connectivity is lost
  res.write('retry: 10000\n\n');

  // Initialize the simulator
  if(simulator == null) {
    // Validate configuration
    var configuration = JSON.parse(fs.readFileSync( __dirname + '/../../data/context/configuration.json'));
    var tcmChannel = configuration.messaging;
    var validateResult = Simulator.validate(tcmChannel);
    if(validateResult == false) {
      res.status(500).write("TCM Kafka channel not defined");
      return;
    }

    // Create and connect the simulator to eFTL
    try {
      simulator = new Simulator.SimulatorConnection(tcmChannel);
      await simulator.connect();
    }
    catch(err) {
      console.log(err);
      return;
    }
  }

  simulator.subscribeStatus(function(status){
    res.write(`data: ${status}\n\n`);
  });

});

  // Start testcase
router.post('/service/simulation/command/:action', async function(req, res, next) {
  // If simulator is not yet initialized, create it now
  if(simulator == null) {
    // Validate TCM configuration
    var configuration = JSON.parse(fs.readFileSync( __dirname + '/../../data/context/configuration.json'));
    var tcmChannel = configuration.messaging;
    if(tcmChannel.url == null || tcmChannel.authKey == null) {
      res.status(500).send("TCM channel not defined");
      return;
    }

    // Create and connect the simulator to Kafka
    try {
      simulator = new Simulator.SimulationCommandPublisher(tcmChannel);
      await simulator.connect();
    }
    catch(err) {
      res.status(500).send(err);
      return;
    }
  }

  // Start a new simulation
  try {
    var action = req.params.action;
    var topic = null;
    if(action == 'command')
      topic = 'iot.simulator.command';
    else if(action == 'status')
      topic = 'iot.simulator.status';
    
    var message = JSON.stringify(req.body);
    simulator.publish(topic, message);
    res.status(200).send('{}');
    return;
  }
  catch(err) {
    console.log(err);
    res.status(500).send(err);
  }
});


// --------------------------------------------------------------------------------
// PROVISIONING
// Get script
router.get('/service/provisioning/script/:scriptName', function(req, res, next) {
  fs.readFile(__dirname + '/../../data/provisioning/kubernetes/' + req.params.scriptName, function (err, data) {
    if (err) {
      throw err; 
    }
    let response = data.toString();
    res.send(response);
  });
});  


module.exports = router;