/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

var debug = require('debug')('iot-webapp:provisioning');
var fs = require('fs');
var properties = require('properties-reader');
var TCDSProvisioning = require('../webapp/provisioning_tcds');
var KubernetesProvisioning = require('../webapp/provisioning_kubernetes');

if(process.env.TEST_MODE != null && process.env.TEST_MODE.trim() === 'true')
  debug("Provisioning is using TEST MODE");

// Class to manage visualization configuration
class Provisioner {
    // Class constructor
    constructor() {
      var config = fs.readFileSync(__dirname + '/../../data/provisioning/configuration.json', 
        {encoding:'utf8', flag:'r'}); 
      this.configuration = JSON.parse(config);
      this.cicProps = properties('src/webapp/cic.properties');
      this.testMode = process.env.TEST_MODE != null && process.env.TEST_MODE.trim() === 'true';
      this.provisionerMap = {};
    }

    // Execute the provisioning process
    async execute(request, messageCallback) {
      // Register the message callback so that progress can be sent back to calling application
      this.messageCallback = messageCallback;

      // Validate that minimum required fields are present on the request
      this.sendProgressMessage("Basic validation");
      if(request.oauth == null)
        throw "OAuth key not provided";
      if(request.region == null)
        throw "Region is not specified";

      try {
        // Validate required fields for each task and build a BOM for execution 
        var tasks = this.buildBOM(request);
        
        // Iterate over each task and execute accordingly
        this.sendProgressMessage("Executing tasks");

        for(var thisTask of tasks) {
          var provisioner = this.provisionerMap[thisTask.type];
          if(provisioner != null)
            await provisioner.executeTask(thisTask);
          else
            throw "Unknown task type " + thisTask.type
        }
        this.sendProgressMessage("Completed provisioning");
        return tasks;
      }
      catch(err) {
        this.sendProgressMessage(err);
        throw err;
      }
    }

    // Send a progress message to the message callback, if there is one
    sendProgressMessage(message) {
      debug("[Provisioning] " + message);
      if(this.messageCallback != null) {
        this.messageCallback(message);
      }
    }

    // Build a BOM for provisioning and validate all required information is available
    buildBOM(request) {
      this.sendProgressMessage("Building BOM");
      var tasks = [];

      // Iterate over the provision types ("groups")
      for(var thisType of request.provisionTypes) {
        var thisGroup = this.configuration.provisioning.types[thisType];

        // Validate the group is defined
        if(thisGroup == null)
          throw "Group " + thisType + " not found in provisioning configuration";

        // Iterate over each task in the group
        for(var thisItem of thisGroup) {
          var thisTask = this.configuration.provisioning.tasks[thisItem];

          // Validate the task is defined
          if(thisTask == null)
            throw "Task " + thisTask + " not found in provisioning configuration";

          // Append minimum required 
          thisTask.region = request.region;
          thisTask.oauth = request.oauth;
                
          // Add task to the return array
          tasks.push(thisTask);
        }
      }

      // Iterate over each of the tasks, and configure appropriately
      for(var thisTask of tasks) {
        this.sendProgressMessage("Building task " + thisTask.type);
        if(thisTask.type == "tcds") {
          var provisioner = new TCDSProvisioning.Provisioner(this);
          this.provisionerMap[thisTask.type] = provisioner;
          provisioner.configureTask(thisTask, request);
        }
        else if(thisTask.type == "kubernetes") {
          var provisioner = new KubernetesProvisioning.Provisioner(this);
          this.provisionerMap[thisTask.type] = provisioner;
          provisioner.configureTask(thisTask, request);
        }
        else
          throw "Unknown task type " + thisTask.type;
      }

      return tasks;
    }
}

module.exports.Provisioner = Provisioner;