/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

//var { ECRClient, GetAuthorizationTokenCommand } = require("@aws-sdk/client-ecr"); // AWS V3
//var { EKSClient, DescribeClusterCommand } = require("@aws-sdk/client-eks"); // AWS V3
//var { STSClient, AssumeRoleCommand } = require("@aws-sdk/client-sts"); // AWS V3
var AWS = require("aws-sdk"); // AWS V2
var axios = require('axios');
var https = require('https');

// Class to manage kubernetes configuration
class Provisioner {
  constructor(provisioner) {
    this.provisioner = provisioner;
    this.cicProps = this.provisioner.cicProps;
    this.testMode = provisioner.testMode;
  }

    // Validate the request has required fields, then set them on the task
    configureTask(task, request) {
      if(request.kafkaBroker == null)
        throw "Kafka broker not provided for Kubernetes task";
      if(request.kafkaUsername == null)
        throw "Kafka username not provided for Kubernetes task";
      if(request.kafkaPassword == null)
        throw "Kafka password not provided for Kubernetes task";
      if(request.awsRegion == null)
        throw "AWS region not provided for Kubernetes task";
      if(request.awsAccessKeyId == null)
        throw "AWS access key id not provided for Kubernetes task";
      if(request.awsSecretAccessKey == null)
        throw "AWS secret access key not provided for Kubernetes task";
      if(request.awsRoleArn == null)
        throw "AWS role ARN not provided for Kubernetes task";
      if(request.dockerImage == null)
        throw "Docker image not provided for Kubernetes task";
      if(request.dockerRegistry == null)
        throw "Docker registry not provided for Kubernetes task";
      if(request.kubernetesClusterName == null)
        throw "Kubernetes cluster name not provided for Kubernetes task";
      if(request.kubernetesToken == null)
        throw "Kubernetes token not provided for Kubernetes task";
  
      task.kafkaBroker = request.kafkaBroker;
      task.kafkaUsername = request.kafkaUsername;
      task.kafkaPassword = request.kafkaPassword;
      task.awsRegion = request.awsRegion;
      task.awsAccessKeyId = request.awsAccessKeyId;
      task.awsSecretAccessKey = request.awsSecretAccessKey;
      task.awsRoleArn = request.awsRoleArn;
      // Ignore these for now
      //task.dockerImage = request.dockerImage;
      //task.dockerRegistry = request.dockerRegistry;
      task.kubernetesClusterName = request.kubernetesClusterName;
      task.kubernetesToken = request.kubernetesToken;
    }

    // Execute the provisioning task
    async executeTask(task) {
      this.sendProgressMessage("Executing Kubernetes task for cluster " + task.kubernetesClusterName);
      if(this.testMode == true) return;

      // Setup AWS credentials object
      var credentials = {
        region: task.awsRegion,
        roleArn: task.awsRoleArn,
        accessKeyId: task.awsAccessKeyId,
        secretAccessKey: task.awsSecretAccessKey
      }

      this.sendProgressMessage("Retrieving AWS cross account credentials");
      var accessParams = await this.getCrossAccountCredentials(credentials);

      this.sendProgressMessage("Waiting for EKS cluster to be ready");
      var waitForEKSClusterResult = await this.waitForEKSCluster(credentials.region, accessParams, task.kubernetesClusterName);

      // Extract certificate and endpoint from EKS
      var eksCertificate = Buffer.from(waitForEKSClusterResult.cluster.certificateAuthority.data, 'base64').toString('ascii');
      var eksEndpoint = waitForEKSClusterResult.cluster.endpoint;
      this.sendProgressMessage("Using EKS endpoint " + eksEndpoint);
      // this.sendProgressMessage("Using EKS certificate " + eksCertificate);

      // Get the ECR token
      this.sendProgressMessage("Retrieving ECR authorization token");
      var ecrGetAuthorizationTokenResult = await this.ecrGetAuthorizationToken(credentials.region, accessParams);
      var ecrAuthorizationToken = ecrGetAuthorizationTokenResult.authorizationData[0].authorizationToken;

      // Create HTTPS agent for secure calls to kubernetes REST services
      this.sendProgressMessage("Creating HTTPS agent for Kubernetes service calls");
      var httpsAgent = new https.Agent({ca: eksCertificate});

      // Create the docker secret
      try {
        this.sendProgressMessage("Creating Kubernetes Docker secret");
        var kubernetesCreateDockerSecretResult = await this.kubernetesCreateDockerSecret(eksEndpoint, task.kubernetesToken, httpsAgent, 
          ecrAuthorizationToken, task.dockerRegistry, task.dockerSecret);
        }
      catch(err) {
        var msg = "Error creating Kubernetes Docker secret: " + err.response.data.message;
        this.handleError(msg);
      }

      // Create the cluster role 'service-update'
      try {
        this.sendProgressMessage("Creating Kubernetes cluster role '" + task.serviceUpdateClusterRole + "'");
        var kubernetesRbacCreateClusterRoleResult = await this.kubernetesRbacCreateClusterRole(eksEndpoint, task.kubernetesToken, httpsAgent, 
            task.serviceUpdateClusterRole);
      }
      catch(err) {
        this.sendProgressMessage("Error creating Kubernetes cluster role: " + err.response.data.message);
      }

      // Bind service account 'default' to cluster role 'service-update'
      try {
        this.sendProgressMessage("Binding Kubernetes service account 'default' to cluster role '" + task.serviceUpdateClusterRole + "'");
        var kubernetesRbacBindDefaultClusterRoleResult = await this.kubernetesRbacBindDefaultClusterRole(eksEndpoint, task.kubernetesToken, httpsAgent, 
            task.serviceUpdateClusterRole);
      }
      catch(err) {
        var msg = "Error binding Kubernetes default service account to cluster role: " + err.response.data.message;
        this.handleError(msg);
      }

      // Create config map for configuration
      try {
        this.sendProgressMessage("Creating Kubernetes configuration map " + task.configMapForConfiguration + " for app " + task.appName);
        var kubernetesCreateConfigMapConfigurationResult = await this.kubernetesCreateConfigMap(eksEndpoint, task.kubernetesToken, httpsAgent, 
            task.appName, task.configMapForConfiguration);
      }
      catch(err) {
        var msg = "Error creating Kubernetes configuration map for configuration: " + err.response.data.message;
        this.handleError(msg);
      }

      // Create config map for resources
      try {
        this.sendProgressMessage("Creating Kubernetes configuration map " + task.configMapForResources + " for app " + task.appName);
        var kubernetesCreateConfigMapResourcesResult = await this.kubernetesCreateConfigMap(eksEndpoint, task.kubernetesToken, httpsAgent, 
            task.appName, task.configMapForResources);
      }
      catch(err) {
        var msg = "Error creating Kubernetes configuration map for resources: " + err.response.data.message;
        this.handleError(msg);
      }

      // Deploy stateful set
      try {
        // Update the substitutions
        var substitutions = task.substitutions
          .replace('${KAFKA_BROKER}', task.kafkaBroker)
          .replace('${KAFKA_USERNAME}', task.kafkaUsername)
          .replace('${KAFKA_PASSWORD}', task.kafkaPassword);
        this.sendProgressMessage("Substituting KAFKA_BROKER " + task.kafkaBroker);
        this.sendProgressMessage("Substituting KAFKA_USERNAME " + task.kafkaUsername);
        this.sendProgressMessage("Substituting KAFKA_PASSWORD " + task.kafkaPassword);
        this.sendProgressMessage("Deploying Kubernetes stateful set");
        var kubernetesDeployStatefulSetResult = await this.kubernetesDeployStatefulSet(eksEndpoint, task.kubernetesToken, httpsAgent, 
              task.appName, task.replicaCount, task.configMapForConfiguration, task.configMapForResources, task.dockerRegistry, task.dockerImage, 
              task.dockerSecret, substitutions);
      }
      catch(err) {
        var msg = "Error deploying Kubernetes stateful set: " + err.response.data.message;
        this.handleError(msg);
      }

      // Retrieve pods
      try {
        this.sendProgressMessage("Retrieving Kubernetes pod");
        await this.waitForKubernetesPod(eksEndpoint, task, httpsAgent);
      }
      catch(err) {
        var msg = "Error retrieving Kubernetes pods: " + err.response.data.message;
        this.handleError(msg);
      }

    }

    // Retrieve cross account credentials
    async getCrossAccountCredentials_V3(credentials) {
        // Create client
      var clientParams = {
        region: credentials.region,
        credentials: {
          accessKeyId: credentials.accessKeyId,
          secretAccessKey: credentials.secretAccessKey
        }
      }
      var client = new STSClient(clientParams);

      // Create command
      var timestamp = (new Date()).getTime();
      var commandParams = {
        RoleArn: credentials.roleArn,
        RoleSessionName: `be-descriptibe-here-${timestamp}`
      };
      var command = new AssumeRoleCommand(commandParams);

      // Invoke
      return await client.send(command);
    };

    async getCrossAccountCredentials(credentials) {
        // Create client
        var sts = new AWS.STS({
          accessKeyId: credentials.accessKeyId,
          secretAccessKey: credentials.secretAccessKey,
          region: credentials.region
        });
        
      // Create command
      var timestamp = (new Date()).getTime();
      var commandParams = {
        RoleArn: credentials.roleArn,
        RoleSessionName: `be-descriptibe-here-${timestamp}`
      };
      
      // Invoke
      var promise = new Promise(function(resolve, reject){
        sts.assumeRole(commandParams, function (err, data) {
          if(err)
            reject(err)
          else
            resolve(data);
        });  
      });

      return promise;
    }

    // Retrieve client params from accessParams
    getClientParams(region, accessParams) {
      var clientParams = {
          region: region,
          credentials: {
              accessKeyId: accessParams.Credentials.AccessKeyId,
              secretAccessKey: accessParams.Credentials.SecretAccessKey,
              sessionToken: accessParams.Credentials.SessionToken
          }
      };
  
      return clientParams;
    };
  
    // Wait for EKS cluster to become active
    async waitForEKSCluster(region, accessParams, clusterName) {
      var self = this;
      var promise = new Promise(async function(resolve, reject){
          // Function to test status of the cluster
          var testStatus = async function() {
              // Describe cluster, keep checking until complete
              var describeClusterResult = await self.eksDescribeCluster(region, accessParams, clusterName);
              self.sendProgressMessage("EKS cluster " + clusterName + " " + describeClusterResult.cluster.status);
              if(describeClusterResult.cluster.status == 'CREATING')
                  setTimeout(testStatus, 10000);
              else if(describeClusterResult.cluster.status == 'ACTIVE')
                  resolve(describeClusterResult);
              else
                  reject(describeClusterResult);
          };

          // Call first iteration
          testStatus();
      });

      return promise;
    };

    // Describe EKS cluster state
    async eksDescribeCluster_V3(region, accessParams, clusterName) {
      // Create client
      var client = new EKSClient(this.getClientParams(region, accessParams));
  
      // Create command
      var commandParams = {
          name: clusterName
      };
      var command = new DescribeClusterCommand(commandParams);
  
      // Invoke
      return await client.send(command);
    };
  
    // Describe EKS cluster state
    async eksDescribeCluster(region, accessParams, clusterName) {
      // Create client
      var eks = new AWS.EKS(this.getClientParams(region, accessParams));
  
      // Create command
      var commandParams = {
          name: clusterName
      };
  
      // Invoke
      var promise = new Promise(function(resolve, reject){
        eks.describeCluster(commandParams, function (err, data) {
          if(err)
            reject(err)
          else
            resolve(data);
        });  
      });

      return promise;
    };
  
    // Get ECR Authorization token
    async ecrGetAuthorizationToken_V3(region, accessParams) {
      // Create client
      var client = new ECRClient(this.getClientParams(region, accessParams));

      // Create command
      var commandParams = {};
      var command = new GetAuthorizationTokenCommand(commandParams);

      // Invoke
      return await client.send(command);
    }

    // Get ECR Authorization token
    async ecrGetAuthorizationToken(region, accessParams) {
      // Create client
      var ecr = new AWS.ECR(this.getClientParams(region, accessParams));

      // Create command
      var commandParams = {};

      // Invoke
      var promise = new Promise(function(resolve, reject){
        ecr.getAuthorizationToken(commandParams, function (err, data) {
          if(err)
            reject(err)
          else
            resolve(data);
        });  
      });

      return promise;
    }

    // Create docker secret for Kubernetes
    async kubernetesCreateDockerSecret(endpoint, token, httpsAgent, ecrAuthorizationToken, dockerRegistry, dockerSecretName) {
      // Decode the ecrAuthorizationToken
      var ecrAuthorizationTokenDecoded = Buffer.from(ecrAuthorizationToken, 'base64').toString('ascii');
      var ecrAuthorizationUserPass = ecrAuthorizationTokenDecoded.split(':');
  
      // Create dockerConfig
      var dockerConfig = {
          auths:{}
      };
      dockerConfig.auths[dockerRegistry] = {
          username: "AWS",
          password: ecrAuthorizationUserPass[1],
          email: "",
          auth: ecrAuthorizationToken
      };
  
      // Convert to JSON and then to Base64
      var dockerConfigJson = Buffer.from(JSON.stringify(dockerConfig), 'ascii').toString('base64');
  
      // Compose request
      var request = {
          "kind": "Secret",
          "apiVersion": "v1",
          "metadata": {
              "name": dockerSecretName,
              "creationTimestamp": null
          },
          "data": {
              ".dockerconfigjson": dockerConfigJson
          },
          "type": "kubernetes.io/dockerconfigjson"
      };
  
      var baseUrl = endpoint;
      var url = '/api/v1/namespaces/default/secrets';
      var headers = {};
      headers['Authorization'] = 'Bearer ' + token;
  
      var axiosResponse = await axios({
          url: url,
          method: 'POST',
          baseURL: baseUrl,
          headers: headers,
          httpsAgent: httpsAgent,
          data: request
        });
  
        return axiosResponse;
    };
  
    // Create a cluster role in RBAC
    async kubernetesRbacCreateClusterRole(endpoint, token, httpsAgent, clusterRole) {
      // Compose request
      var request = {
          "kind": "ClusterRole",
          "apiVersion": "rbac.authorization.k8s.io/v1",
          "metadata": {
              "name": clusterRole,
              "namespace": "default",
              "creationTimestamp": null
          },
          "rules": [
              {
                  "verbs": [
                      "get",
                      "list",
                      "watch",
                      "create",
                      "update",
                      "patch",
                      "delete"
                  ],
                  "apiGroups": [
                      ""
                  ],
                  "resources": [
                      "services",
                      "pods"
                  ]
              }
          ]
      };
  
      var baseUrl = endpoint;
      var url = '/apis/rbac.authorization.k8s.io/v1/clusterroles';
      var headers = {};
      headers['Authorization'] = 'Bearer ' + token;
  
      var axiosResponse = await axios({
          url: url,
          method: 'POST',
          baseURL: baseUrl,
          headers: headers,
          httpsAgent: httpsAgent,
          data: request
        });
  
        return axiosResponse;
    };
  
    // Bind the cluster role to the default user in RBAC
    async kubernetesRbacBindDefaultClusterRole(endpoint, token, httpsAgent, clusterRole) {
      // Compose request
      var request = {
          //"kind": "RoleBinding",
          "kind": "ClusterRoleBinding",
          "apiVersion": "rbac.authorization.k8s.io/v1",
          "metadata": {
              "name": clusterRole,
              "namespace": "default",
              "creationTimestamp": null
          },
          "subjects": [
              {
                  "kind": "User",
                  "apiGroup": "rbac.authorization.k8s.io",
                  "name": "system:serviceaccount:default:default"
              }
          ],
          "roleRef": {
              "apiGroup": "rbac.authorization.k8s.io",
              "kind": "ClusterRole",
              "name": clusterRole
          }
      };
  
      var baseUrl = endpoint;
      var url = '/apis/rbac.authorization.k8s.io/v1/clusterrolebindings';
      var headers = {};
      headers['Authorization'] = 'Bearer ' + token;
  
      var axiosResponse = await axios({
          url: url,
          method: 'POST',
          baseURL: baseUrl,
          headers: headers,
          httpsAgent: httpsAgent,
          data: request
        });
  
        return axiosResponse;
    };
  
    // Create a configuration map
    async kubernetesCreateConfigMap(endpoint, token, httpsAgent, appName, configName) {
      var request = {
          "kind": "ConfigMap",
          "apiVersion": "v1",
          "metadata": {
              "name": configName,
              "creationTimestamp": null,
              "labels": {
                  "app.kubernetes.io/name": configName,
                  "app.kubernetes.io/part-of": appName
              }
          }
      };
  
      var baseUrl = endpoint;
      var url = '/api/v1/namespaces/default/configmaps?fieldManager=kubectl-client-side-apply';
      var headers = {};
      headers['Authorization'] = 'Bearer ' + token;
  
      var axiosResponse = await axios({
          url: url,
          method: 'POST',
          baseURL: baseUrl,
          headers: headers,
          httpsAgent: httpsAgent,
          data: request
        });
  
        return axiosResponse;
    };
  
    // Deploy a stateful ste
    async kubernetesDeployStatefulSet(endpoint, token, httpsAgent, appName, replicaCount, configMapForConfiguration, 
        configMapForResources, dockerRegistry, dockerImage, dockerSecret, substitutions) {
      var request = {
          "kind": "StatefulSet",
          "apiVersion": "apps/v1",
          "metadata": {
              "name": appName,
              "creationTimestamp": null,
              "labels": {
                  "app.kubernetes.io/name": appName,
                  "app.kubernetes.io/part-of": appName
              }
          },
          "spec": {
              "replicas": replicaCount,
              "selector": {
                  "matchLabels": {
                      "app": appName
                  }
              },
              "template": {
                  "metadata": {
                      "creationTimestamp": null,
                      "labels": {
                          "app": appName,
                          "app.kubernetes.io/part-of": appName
                      }
                  },
                  "spec": {
                      "volumes": [
                          {
                              "name": "configuration",
                              "configMap": {
                                  "name": configMapForConfiguration,
                                  "defaultMode": 420
                              }
                          },
                          {
                              "name": "resources",
                              "configMap": {
                                  "name": configMapForResources,
                                  "defaultMode": 420
                              }
                          }
                      ],
                      "containers": [
                          {
                              "name": appName,
                              "image": dockerRegistry + "/" + dockerImage,
                              "env": [
                                  {
                                      "name": "POD_NAME",
                                      "valueFrom": {
                                          "fieldRef": {
                                              "apiVersion": "v1",
                                              "fieldPath": "metadata.name"
                                          }
                                      }
                                  },
                                  {
                                      "name": "POD_NAMESPACE",
                                      "valueFrom": {
                                          "fieldRef": {
                                              "apiVersion": "v1",
                                              "fieldPath": "metadata.namespace"
                                          }
                                      }
                                  },
                                  {
                                      "name": "APPNAME",
                                      "value": appName
                                  },
                                  {
                                      "name": "STREAMING_NODENAME",
                                      "value": "$(POD_NAME).$(APPNAME)"
                                  },
                                  {
                                      "name": "STREAMING_SUBSTITUTIONS",
                                      "value": substitutions
                                  }
                              ],
                              "resources": {},
                              "volumeMounts": [
                                  {
                                      "name": "configuration",
                                      "mountPath": "/var/opt/tibco/streambase/configuration"
                                  },
                                  {
                                      "name": "resources",
                                      "mountPath": "/var/opt/tibco/streambase/resources"
                                  }
                              ],
                              "livenessProbe": {
                                  "httpGet": {
                                      "path": "/healthcheck/v1/status",
                                      "port": 8008,
                                      "scheme": "HTTP"
                                  },
                                  "initialDelaySeconds": 240,
                                  "timeoutSeconds": 1,
                                  "periodSeconds": 10,
                                  "successThreshold": 1,
                                  "failureThreshold": 3
                              },
                              "readinessProbe": {
                                  "httpGet": {
                                      "path": "/healthcheck/v1/status",
                                      "port": 8008,
                                      "scheme": "HTTP"
                                  },
                                  "initialDelaySeconds": 60,
                                  "timeoutSeconds": 1,
                                  "periodSeconds": 10,
                                  "successThreshold": 1,
                                  "failureThreshold": 3
                              },
                              "terminationMessagePath": "/dev/termination-log",
                              "terminationMessagePolicy": "File",
                              "imagePullPolicy": "IfNotPresent",
                              "tty": true
                          }
                      ],
                      "restartPolicy": "Always",
                      "terminationGracePeriodSeconds": 30,
                      "dnsPolicy": "ClusterFirst",
                      "securityContext": {},
                      "imagePullSecrets": [
                          {
                              "name": dockerSecret
                          }
                      ],
                      "schedulerName": "default-scheduler"
                  }
              },
              "serviceName": appName,
              "podManagementPolicy": "Parallel",
              "updateStrategy": {
                  "type": "RollingUpdate"
              },
              "revisionHistoryLimit": 10
          },
          "status": {
              "replicas": 0
          }
      };

      var baseUrl = endpoint;
      var url = '/apis/apps/v1/namespaces/default/statefulsets?fieldManager=kubectl-client-side-apply';
      var headers = {};
      headers['Authorization'] = 'Bearer ' + token;

      var axiosResponse = await axios({
          url: url,
          method: 'POST',
          baseURL: baseUrl,
          headers: headers,
          httpsAgent: httpsAgent,
          data: request
        });

        return axiosResponse;
    };
   
    // Wait for a kubernetes pod to be running
    async waitForKubernetesPod(eksEndpoint, task, httpsAgent) {
      var self = this;
      var getPodsCount = 0;
      var podCountLimit = 10;
      var waitDelay = 2000;

      var promise = new Promise(async function(resolve, rejct){
        // Function to call get pods
        var callGetPods = async function() {
          var kubernetesGetPodsResult = await self.kubernetesGetPods(eksEndpoint, task.kubernetesToken, httpsAgent, task.namespace);
          getPodsCount++;

          // If no data is returned, timeout for a second and try again up to 5 times
          if(getPodsCount <= podCountLimit && (kubernetesGetPodsResult.data == null || kubernetesGetPodsResult.data.items == null || 
            kubernetesGetPodsResult.data.items.length == 0)) {
              self.sendProgressMessage("Kubernetes pod starting");
              setTimeout(callGetPods, waitDelay);
          }
          // If less than 5 times
          else if(getPodsCount < podCountLimit) {
            var item = kubernetesGetPodsResult.data.items[0];
            self.sendProgressMessage("Kubernetes pod " + item.metadata.name + " status " + item.status.phase);

            // If not running, timeout for a second and try again up to the number of times
            if(item.status.phase != 'Running') {
              setTimeout(callGetPods, waitDelay);
            }
            // Otherwise we're done
            else {
              resolve();
            }
          }
          // Otherwise probably something wrong, but send a message
          else {
            self.sendProgressMessage("Kubernetes pod still starting, check status manually later");
            resolve();
          }
        }

        // Call first get pods
        callGetPods();
      })

      return promise;
    };

    // Get pods for a kubernetes cluster
    async kubernetesGetPods(endpoint, token, httpsAgent, namespace) {
      var baseUrl = endpoint;
      var url = '/api/v1/namespaces/' + namespace + '/pods?allowWatchBookmarks=true&limit=26878832&timeoutSeconds=26878832';
      var headers = {};
      headers['Authorization'] = 'Bearer ' + token;
      
      var axiosResponse = await axios({
          url: url,
          method: 'GET',
          baseURL: baseUrl,
          headers: headers,
          httpsAgent: httpsAgent
        });
  
        return axiosResponse;
    };
  
    // Send a progress message
    sendProgressMessage(message) {
      this.provisioner.sendProgressMessage(message);
    }

    // Handle an error
    handleError(msg) {
      // This will ignore messages where already exists, otherwise throw exception to stop provisioning
      if(msg.includes('already exists'))
        this.sendProgressMessage(msg);
      else
        throw msg
    }
}

module.exports.Provisioner = Provisioner;