/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

var axios = require('axios');
var fs = require('fs');
var archiver = require('archiver');
var properties = require('properties-reader');

// Class to manage spotfire dxp configuration
class Provisioner {
    constructor() {
      var config = fs.readFileSync(__dirname + '/../../data/provisioning/configuration.json', 
        {encoding:'utf8', flag:'r'}); 
      this.configuration = JSON.parse(config);
      this.cicProps = properties('src/webapp/cic.properties');
    }
  
    // Lookup stream names from TCDS
    async lookupStreams(region, authorization) {
        var self = this;
        var promise = new Promise(function(resolve, reject){
            var url = '/api/v1/streams/all/runtime';
            var baseUrl = self.cicProps.get(region + '.TIBCO_CLOUD_DATASTREAMS');
        
            axios({
                url: url,
                method: 'GET',
                baseURL: baseUrl,
                headers: {
                    'Authorization': authorization
                }
            })
            .then(function(response){
                resolve(response.data.streams);
            })
            .catch(function(err){
                reject(err);
            });
                
        });

        return promise;
    };

    // Write data access plan file with substituted variables
    async writeDataAccessPlan(region, task) {
        // Read the template file
        var dataAccessPlanTemplateFileName = __dirname + task.location + task.sourcePath + task.dataAccessPlanTemplateFileName;
        var dataAccessPlanTemplate = fs.readFileSync(dataAccessPlanTemplateFileName, {encoding:'utf8', flag:'r'});

        var dataAccessPlan = dataAccessPlanTemplate;
        for(var thisStream of task.streams) {
            var url = this.cicProps.get(region + '.TIBCO_CLOUD_DATASTREAMS') + ':443/api/v1/streams/' + thisStream.id + '/runtime/connect'
            var regex = new RegExp('\\${' + thisStream.param + '}', 'g');
            dataAccessPlan = dataAccessPlan.replace(regex, url);
        };

        // Update the data access plan
        var dataAccessPlanFileName = __dirname + task.location + task.sourcePath + task.dataAccessPlanFileName;
        fs.writeFileSync(dataAccessPlanFileName, dataAccessPlan);
    };

    // Download DXP file
    async download(region, authorization, response) {
        // Get the task
        var task = this.configuration.provisioning.tasks.spotfire;

        // Lookup the streams
        var streams = await this.lookupStreams(region, authorization);
        var streamMap = {};
        for(var thisStream of streams) {
            streamMap[thisStream.name] = thisStream;
        }

        // Map the id for the stream
        for(var thisStream of task.streams) {
            var stream = streamMap[thisStream.name];
            if(stream == null) {
                response.status(500).send("Stream " + thisStream.name + " not found");
                return;
            }
            thisStream.id = stream.id;
        }

        await this.writeDataAccessPlan(region, task);

        // Create the zip
        var zip = archiver('zip', task.zipOptions);

        // Set the download name
        response.attachment(task.dxp);

        await new Promise(function(resolve, reject) {
            // Pipe the archive to the response
            zip.pipe(response);

            // Append the directory of files
            zip.directory(__dirname + task.location + task.sourcePath, false);

            // Error handler
            zip.on('error', function(err){
                throw err;
            });

            // Finalize the archive
            zip.finalize();

            // When the stream is finished
            response.on('close', function() {
                //response.send('OK').end();
                response.end();
                resolve();
            });
        }); 
    };
}

module.exports.Provisioner = Provisioner;