/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

var axios = require('axios');
var fs = require('fs');

// Class to manage TCDS configuration
class Provisioner {
    constructor(provisioner) {
      this.provisioner = provisioner;
      this.cicProps = this.provisioner.cicProps;
      this.testMode = provisioner.testMode;
    }

    // Validate the request has required fields for TCDS, then set them on the task
    configureTask(task, request) {
      if(request.kafkaBroker == null)
        throw "Kafka broker url not provided for TCDS task";
      if(request.kafkaUsername == null)
        throw "Kafka username not provided for TCDS task";
      if(request.kafkaPassword == null)
        throw "Kafka password not provided for TCDS task";

      task.kafkaBroker = request.kafkaBroker;
      task.kafkaUsername = request.kafkaUsername;
      task.kafkaPassword = request.kafkaPassword;
    }

    // Execute the TCDS task
    async executeTask(task) {
      // Retrieve streams the first time only
      if(this.streamMap == null) {
        var streams = await this.lookupStreams(task);
        this.streamMap = {};
        for(var thisStream of streams) {
            this.streamMap[thisStream.name] = thisStream;
        }
      }

      // Validate if stream already exists
      if(this.streamMap[task.streamName] != null) {
        var stream = this.streamMap[task.streamName];
        this.sendProgressMessage("TCDS stream " + task.streamName + " already exists");
        task.streamId = stream.id;
        if(stream.status == 'Stopped') {
          await this.startStream(task);
        }
        return;
      }

      this.sendProgressMessage("Executing TCDS task for " + task.streamTemplate);

      // Read the template file
      var templateFileName = __dirname + task.location + "/" + task.streamTemplate;
      var templateDefinition = fs.readFileSync(templateFileName, {encoding:'utf8', flag:'r'});

      // Update the template
      var streamDefinition = templateDefinition
        .replace('${KAFKA_BROKER}', task.kafkaBroker)
        .replace('${KAFKA_USERNAME}', task.kafkaUsername)
        .replace('${KAFKA_PASSWORD}', task.kafkaPassword);

        try {
          await this.pushStream(task, streamDefinition);
          await this.startStream(task);
          this.sendProgressMessage("TCDS task for " + task.streamTemplate + " completed");
        }
        catch(err) {
          throw "TCDS task for " + task.streamTemplate + " failed: " + err;
        }
    }

    // Lookup streams
    async lookupStreams(task) {
      // Compose urls
      var url = '/api/v1/streams/all/runtime';
      var baseUrl = this.cicProps.get(task.region + '.TIBCO_CLOUD_DATASTREAMS');

      this.sendProgressMessage("Retrieving existing TCDS streams");
      try {
        var response = await axios({
              url: url,
              method: 'GET',
              baseURL: baseUrl,
              headers: {
                'Authorization': 'Bearer ' + task.oauth
              }
            });
        return response.data.streams;
      }
      catch(err) {
        throw "Lookup TCDS Streams: " + err.response.status + " " + err.response.statusText; 
      }
    };
  
    // Push TCDS stream definition to the server
    async pushStream(task, streamDefinition) {
      this.sendProgressMessage("Pushing TCDS " + task.streamName + " to " + task.region + " region");
      if(this.testMode == true) return;

      // Compose urls
      var url = '/api/v1/streams';
      var baseUrl = this.cicProps.get(task.region + '.TIBCO_CLOUD_DATASTREAMS');

      var headers = {};
      headers['Authorization'] = 'Bearer ' + task.oauth;

      // Call provisioning service 
      try {     
        var response = await axios({
            url: url,
            method: 'POST',
            baseURL: baseUrl,
            headers: headers,
            data: JSON.parse(streamDefinition)
          });

        task.streamId = response.data.id;
      }
      catch(err) {
        throw "Push TCDS Stream: " + err.response.status + " " + err.response.statusText; 
      }
    }

    // Start TCDS stream
    async startStream(task) {
      this.sendProgressMessage("Starting TCDS stream " + task.streamName);
      if(this.testMode == true) return;

      // Compose urls
      var url = '/api/v1/streams/' + task.streamId + ':start';
      var baseUrl = this.cicProps.get(task.region + '.TIBCO_CLOUD_DATASTREAMS');

      var headers = {};
      headers['Authorization'] = 'Bearer ' + task.oauth;
      
      var body = {
        "id": task.streamId
      }

      // Call start service
      try {
        await axios({
            url: url,
            method: 'POST',
            baseURL: baseUrl,
            headers: headers,
            data: body
          });
      }
      catch(err) {
        throw "Start TCDS Stream " + err.response.status + " " + err.response.statusText; 
      }
    };

    // Send a progress message
    sendProgressMessage(message) {
      this.provisioner.sendProgressMessage(message);
    }

}

module.exports.Provisioner = Provisioner;