/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

var eFTL = require('../../lib/eftl');
var debug = require('debug')('iot-webapp:simulator');

// Class to manage simulation
class Simulator {
    constructor(tcmChannel) {
        this.tcmChannel = tcmChannel;
        this.sims = [];
    }

    // Connect to TCM
    async connect() {
        var self = this;
        var promise = new Promise(function(resolve, reject){
            // Create connection to EFTL
            self.eftl_connection= null;

            debug("Connecting to eFTL " + self.tcmChannel.url);
            eFTL.connect(self.tcmChannel.url, {
                password: self.tcmChannel.authKey,
                clientId: "iot-webapp",
                onConnect: function(connection) {
                    self.eftl_connection = connection;
                    debug("Connection success");
                    resolve();
                },
                onDisconnect: function(connection, code, reason) {
                    self.eftl_connection = null;
                    debug("Connection failed: " + code + ": " + reason);
                    reject(code + " " + reason);
                }
            });
        });

        return promise;
    }

    formatDateTime(date) {

    }

    generateBarcode(length) {
        var result = '';
        var characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
        var charactersLength = characters.length;
        for(var i = 0; i < length; i++) {
            result += characters.charAt(Math.floor(Math.random() * charactersLength));
        }
        return "BC" + result;
    }

    // Start a new simulation
    async simulate(testcase, compression) {
        var self = this;
        var promise = new Promise(function(resolve, reject){
            var barcode = self.generateBarcode(14);
            var baseDateTime = new Date(Date.UTC(2020, 1 /* Feb */, 12, 14, 0, 2, 0));
            var observations = [];
    
            // Create reports
            var announcement = {
                report: testcase.data.announcement
            }
            announcement.report.eventDateTime = baseDateTime.toISOString().replace('.000Z', 'Z');
            announcement.report.barcode = barcode;
            
            // Create observations
            for(var idx = 0; idx < testcase.data.observations.length; idx++) {
                var thisObservation = testcase.data.observations[idx];
                if(thisObservation.barcode == null)
                    thisObservation.barcode = barcode;
                var tmp = new Date(baseDateTime.getTime() + (thisObservation.offset + new Date().getTimezoneOffset()) * 60000);
                thisObservation.eventDateTime = (new Date(Date.UTC(tmp.getFullYear(), tmp.getMonth(), tmp.getDate(), tmp.getHours(), tmp.getMinutes(), tmp.getSeconds()))).toISOString().replace('.000Z', 'Z');
                delete thisObservation.offset;
                var observation = {
                    report:thisObservation
                }
                observations.push(observation);
            }

            // Create the simulation object
            var simulation = new Simulation(self, announcement, observations, compression, function() {
                simulation = null;
            });
            simulation.execute();

            // Return the barcode
            resolve(barcode);
        });

        return promise;
    }

    async publishEFTLMessage(dest, payload) {
        // Create eftl message
        if(this.eftl_connection == null)
            await this.connect();
        var message = this.eftl_connection.createMessage();
        message.set("_dest", dest);
        message.set("payload", payload);

        // Publish eftl message
        this.eftl_connection.publish(message, {
            onComplete: function(message) {
            },
            onError: function(message, code, reason) {
                debug(reason);
            }
        });
    };
}

class Simulation {
    constructor(simulator, announcement, observations, compression, onComplete) {
        this.simulator = simulator
        this.announcement = announcement;
        this.observations = observations;
        this.compression = compression;
        this.onComplete = onComplete;
    }

    registerObservation() {        
        // If no more observations, just return
        if(this.observations.length == 0) {
            if(this.onComplete != null)
                this.onComplete();
            return;
        }

        // Pop the next observation
        var delay = this.compression * 1000;
        var observation = this.observations[0];
        this.observations.shift();

        // Set timeout the publish
        var self = this;
        setTimeout(function() {
            self.simulator.publishObservation(observation);
            self.registerObservation();
        }, delay)
    };


    async execute() {
        // Publish the announcement
        await this.simulator.publishAnnouncement(this.announcement);

        // Register the first observation
        this.registerObservation();
    }
}

class SimulatorConnection {
    constructor(tcmChannel) {
        this.tcmChannel = tcmChannel;
    }

    // Connect to TCM
    async connect() {
        var self = this;
        var promise = new Promise(function(resolve, reject){
            // Create connection to EFTL
            self.eftl_connection= null;

            debug("Connecting to eFTL " + self.tcmChannel.url);
            eFTL.connect(self.tcmChannel.url, {
                password: self.tcmChannel.authKey,
                clientId: "iot-webapp",
                onConnect: function(connection) {
                    self.eftl_connection = connection;
                    debug("Connection success");
                    resolve();
                },
                onDisconnect: function(connection, code, reason) {
                    self.eftl_connection = null;
                    debug("Connection failed: " + code + ": " + reason);
                    reject(code + " " + reason);
                }
            });
        });

        return promise;
    }

    // Publish command to simulator
    async publish(dest, action, payload) {
        // If not connected, connect now
        if(this.eftl_connection == null)
            await this.connect();

        // Create eftl message
        var message = this.eftl_connection.createMessage();
        message.set("_dest", dest);
        message.set("action", action);
        message.set("payload", payload);

        // Publish eftl message
        this.eftl_connection.publish(message, {
            onComplete: function(message) {
            },
            onError: function(message, code, reason) {
                debug(reason);
            }
        });
    };

    // Subscribe to status messages
    async subscribeStatus(statusCallback) {
        // If not connected, connect now
        if(this.eftl_connection == null)
            await this.connect();

        var self = this;
        var promise = new Promise(function(resolve, reject){
            self.eftl_connection.subscribe({
                matcher: '{"_dest": "iot.simulator.status"}',
                onMessage: function(message) {
                    statusCallback(message);
                }
            });
        });

        return promise;
    }

}

module.exports.SimulatorConnection = SimulatorConnection;
