/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

const { Kafka, logLevel } = require('kafkajs');
var debug = require('debug')('iot-webapp:simulator_kafka');

class SimulatorConnection {
    constructor(tcmChannel) {
        this.tcmChannel = tcmChannel;
    }

    // Connect to Kafka
    async connect() {
        var self = this;
        var promise = new Promise(async function(resolve, reject){
            // Create Kafka client
            var kafkaConfig = {
                clientId: 'IOTSimulatorConnection-' + Date.now(),
                brokers: [self.tcmChannel.kafkaBroker],
                //logLevel: logLevel.DEBUG,
                ssl: true,
                sasl: {
                    mechanism: 'plain',
                    username: self.tcmChannel.kafkaUsername,
                    password: self.tcmChannel.kafkaPassword
                },
                retry: {
                    retries: 0
                }
            }

            self.kafka_client = new Kafka(kafkaConfig);

            // kafkajs seems to have a bug in 1.15.0 where if the group coordinator
            //   and broker URLs are different, then it goes into a connect retry 
            //   loop due to the broker not being found in metadata. If the
            //   group coordinator and broker happen to be the same URL, then it 
            //   works OK, but this can't be relied on to happen.
            // This has been reported to the KafkaJS team on github as an issue.
            // With TCM Kafka normally the cluster consistes of two nodes. 
            //   Running this admin describeCluster command *seems* to solve the
            //   problem by caching both of the nodes in library metadata at this
            //   time. Fingers crossed this keeps working.
            var admin = self.kafka_client.admin();
            await admin.connect();
            //debug('Describing Kafka cluster');
            var describe = await admin.describeCluster();
            //debug(describe);
            await admin.disconnect();

            resolve();

        });
        return promise;
    }

    // Publish command to simulator
    async publish(topic, message) {
        // If not connected, connect now
        if(this.kafka_client == null)
            await this.connect();

        // Get the Kafka producer
        if(this.kafka_producer == null) {
            this.kafka_producer = this.kafka_client.producer();
            await this.kafka_producer.connect();
        }

        // Send message
        await this.kafka_producer.send({
            topic: topic,
            messages: [{key: 'simulation', value: message}]
        });
    };

    // Subscribe to status messages
    async subscribeStatus(statusCallback) {
        // If not connected, connect now
        if(this.kafka_client == null)
            await this.connect();

        var self = this;
        var promise = new Promise(async function(resolve, reject){
            var topic = 'iot.simulator.livestatus';
            var consumer = self.kafka_client.consumer({ groupId: 'IOTSimulatorControlStatus'});
            
            debug("Connecting to Kafka " + self.tcmChannel.kafkaBroker + " for simulation status");
            await consumer.connect();
            
            debug("Subscribing to simulation status topic " + topic);
            await consumer.subscribe({
                topic: topic
            });

            await consumer.run({
                eachMessage: async function({topic, partition, message}) {
                    statusCallback(message.value.toString());
                }
            });

            debug("Completed subscribing to simulation status topic " + topic);
            resolve();
        });

        return promise;
    }

}

// Validates if connection details are adequate for this Simulator
function validate(tcmChannel) {
    return tcmChannel.kafkaBroker != null && tcmChannel.kafkaUsername != null && tcmChannel.kafkaPassword != null; 
}

module.exports.validate = validate;
module.exports.SimulatorConnection = SimulatorConnection;
